# coding=utf-8
"""
FAAPIService
"""
import asyncio
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime, timedelta, tzinfo
from inspect import stack as i_stack
from json import dumps, loads, JSONDecodeError
from time import time
from traceback import format_exc
from typing import Any, Callable, Coroutine, Dict, List, Optional, Tuple, Union

import aiohttp
import bs4
from viledatools import FARequests, getcookieval


@dataclass
class HTTPClientSessionStatus:
    login_mutex: Optional[bool] = None
    login_timestamp: Optional[float] = None


class FAAPIService:
    """
    Base class to build FA API services
    """

    def __init__(self,
                 log: Callable[..., Coroutine[Any, Any, Any]],
                 vapp: str,
                 uagent: str,
                 webrelogin: int,
                 graphql: Dict[str, str] = None,
                 debuglog: bool = False):
        """
        Init the base class

        :param log: logger callable coroutine to use
        :param vapp: FA mobile API app version to use
        :param uagent: User-Agent string to use for mobile API
        :param uagent: web API relogin interval
        :param graphql: GraphQL requests, key is GraphQL method name, value is request body
        :param debuglog: enable debug logging
        """
        # Reference logging object of the caller
        self._base_log: Callable[..., Coroutine[Any, Any, Any]] = log
        # FA mobile API version
        self._vapp = vapp
        # FA mobile API standard headers to provide
        self._app_headers = {"version":    vapp,
                             "Accept":     "*/*",
                             "User-Agent": uagent
                             }
        # Web relogin interval in seconds
        self._webrelogin = float(webrelogin)
        # GraphQL requests
        self._graphql = dict()
        if isinstance(graphql, dict):
            self._graphql = graphql
        else:
            self._graphql = dict()
        # Debug logging
        self._debuglog = debuglog
        # Multiple client sessions
        self._s: Dict[str, aiohttp.ClientSession] = dict()
        self._s_status: Dict[str, HTTPClientSessionStatus] = dict()
        # FA web api time reg draw counter
        self._fa_api_draw = 1

    @staticmethod
    def fa_extract_recurrence(recurrence: Dict[str, Any],
                              taskid: int) -> List[Dict[str, Any]]:
        """
        Extracts task sequence by date from response of FA mobile API tasks request

        :param recurrence: contents of FA mobile API tasks request "recurrence"
        :param taskid: FA backend task ID
        :return: corresponding task sequence
        """
        _res = list()
        for _task_str_id in recurrence.keys():
            if int(_task_str_id) == taskid:
                for _seq_nr_str, _seq_dates_list in recurrence[_task_str_id].items():
                    _res.append({"sequenceNumber": int(_seq_nr_str),
                                 "_date_start":    _seq_dates_list[0],
                                 "_date_end":      _seq_dates_list[1]
                                 })
                return _res
        return _res

    @staticmethod
    def parse_form_html(rawhtml: str) -> Dict[str, Any]:
        """
        Parses the raw FA form submission questions page HTML into structured dict

        :param rawhtml: raw string with HTML contents
        :result: dictionary with submitted form contents
        """
        _strainer_content = bs4.SoupStrainer(name="content", id="Content")
        _bs_content = bs4.BeautifulSoup(rawhtml, features="lxml", parse_only=_strainer_content)
        _out = dict()
        for _child in _bs_content.content.children:
            if isinstance(_child,
                          bs4.element.Tag) and _child.name == "div" and "card" in _child["class"]:
                if "card-header" in _child.div["class"]:
                    _category_name = _child.div.contents[0]
                else:
                    _category_name = None
                _category_fields = list()
                if "card-body" in _child.ul["class"]:
                    for _ul_child in _child.ul:
                        _field = dict()
                        if isinstance(_ul_child,
                                      bs4.element.Tag) and _ul_child.name == "li" and "list-group-item" in _ul_child[
                                                                                                            "class"]:
                            _field["name"] = _ul_child.label.contents[0].replace(chr(10), '').strip()
                            _field["question"] = int(_ul_child.label["for"].split("question-")[1])
                            if _ul_child.div is not None:
                                _field_img = _ul_child.div.find(name="img")
                                _field_lookup = _ul_child.div.find_all(name="div", class_="lookup-table-answers")
                                if _field_img is not None and not _field_lookup:
                                    _field["type"] = "image"
                                    _field["val"] = _field_img["src"]
                                elif _field_img is None and not _field_lookup:
                                    _field["type"] = "value"
                                    _field["val"] = _ul_child.div.contents[0].replace(chr(10), '').strip()
                                elif _field_lookup:
                                    _field["type"] = "lookup"
                                    _selections = list()
                                    for _lookup in _field_lookup:
                                        _selection_item = dict()
                                        _lookup_selection = _lookup.find(name="div", class_="lookup-table-selection")
                                        _selection_item_key = _lookup_selection.contents[0].replace(chr(10), '').strip()
                                        _lookup_subanswers = _lookup.find(name="div", class_="sub-answers")
                                        # Parse subanswers into tuples
                                        _sa_names = list()
                                        _sa_vals = list()
                                        for _sa_child in _lookup_subanswers:
                                            if isinstance(_sa_child, bs4.element.Tag):
                                                if _sa_child.name == "label":
                                                    _sa_names.append(_sa_child.contents[0].replace(chr(10), '').strip())
                                                elif _sa_child.name == "div":
                                                    if _sa_child.contents:
                                                        try:
                                                            _sa_val = int(
                                                                _sa_child.contents[0].replace(chr(10), '').strip())
                                                        except ValueError:
                                                            _sa_val = _sa_child.contents[0].replace(chr(10), '').strip()
                                                        _sa_vals.append(_sa_val)
                                                    else:
                                                        _sa_vals.append(None)
                                        _lookup_subanswers_tuples = list()
                                        if len(_sa_names) == len(_sa_vals):
                                            for _i in range(len(_sa_names)):
                                                _lookup_subanswers_tuples.append((_sa_names[_i], _sa_vals[_i]))
                                        else:
                                            # Parsing exception
                                            pass
                                        _selection_item[_selection_item_key] = _lookup_subanswers_tuples
                                        _selections.append(_selection_item)
                                    _field["val"] = _selections
                            if _field:
                                _category_fields.append(_field)
                _out[_category_name] = _category_fields
        return _out

    async def faweb_login(self, domain: str, usr: str, pwd: str) -> bool:
        """
        Login to FA web API, keep session, and return True if successful

        :param domain: FA domain name
        :param usr: FA user login name
        :param pwd: FA password
        :return: True if successful, None or False if not
        """
        _s_name = self._faweb_get_session_name(domain, usr)
        if self._s_status[_s_name].login_mutex:
            return True
        else:
            self._s_status[_s_name].login_mutex = True
        try:
            r = await self._s[_s_name].head(f"https://{domain}.facilityapps.com/login")
            await self._base_log(i_stack(),
                                 "debug",
                                 (f"OK got init XSRF-TOKEN cookie = "
                                  f"{getcookieval(self._s[_s_name], 'XSRF-TOKEN')} from {r.url.path}"))
            r = await self._s[_s_name].post(f"https://{domain}.facilityapps.com/login",
                                            data={"_token":   getcookieval(self._s[_s_name], "XSRF-TOKEN"),
                                                  "username": usr,
                                                  "password": pwd,
                                                  "submit":   str()
                                                  })
            if getcookieval(self._s[_s_name], "laravel_token"):
                # Successfully authenticated
                self._s_status[_s_name].login_mutex = False
                self._s_status[_s_name].login_timestamp = time()
                # Log the OK
                await self._base_log(i_stack(), "info", (f"OK logged in to {r.url.path} -> {r.content_type} "
                                                         f"[{r.content_length}B] redirects counted: {len(r.history)}"))
                return True
            else:
                self._s_status[_s_name].login_mutex = False
                await self._base_log(i_stack(), "error", (f"FAILED to log in web API, redirected to {r.url.path} "
                                                          f"redirects counted: {len(r.history)}"))
                return False
        except aiohttp.ClientResponseError as _badstatus:
            self._s_status[_s_name].login_mutex = False
            await self._base_log(i_stack(), "debug", f"Full failed URL: {_badstatus.request_info.url}")
            await self._base_log(i_stack(), "error", f"Error {_badstatus.request_info.method} "
                                                     f"{_badstatus.request_info.url.path} - "
                                                     f"{_badstatus.status} {_badstatus.message}")

    async def faweb_logout(self, domain: str, usr: str) -> bool:
        """
        Logout from FA web API for all accounts
        """
        _s_name = self._faweb_get_session_name(domain, usr)
        while self._s_status[_s_name].login_mutex:
            await asyncio.sleep(1.000)
        r = await self._s[_s_name].get(f"https://{domain}.facilityapps.com/logout")
        if r.ok:
            await self._base_log(i_stack(),
                                 "info",
                                 f"OK web logout successful for user '{usr}' from domain '{domain}'")
            _res = True
        else:
            await self._base_log(i_stack(),
                                 "info",
                                 f"FAILED web logout for user '{usr}' from domain '{domain}'")
            _res = False
        await self._s[_s_name].close()
        return _res

    async def faweb_get_configview(self,
                                   domain: str,
                                   usr: str) -> Optional[Dict[str, Any]]:
        """
        Get the config parameters for enfironment for FA API domain

        :param domain: FA domain name
        :param usr: FA user login name
        :return: dict mapping user ID: site object
        """
        # Output dict
        _c = dict()
        _s_name = self._faweb_get_session_name(domain, usr)
        while self._s_status[_s_name].login_mutex:
            await asyncio.sleep(1.000)
        r = await self._s[_s_name].get(f"https://{domain}.facilityapps.com/system_configuration/view",
                                       headers={"Accept": "text/html, application/xhtml+xml"})
        if r.ok:
            _raw_html = await r.text()
        else:
            await self._base_log(i_stack(), "error",
                                 f"FAILED requesting system config for domain '{domain}' user '{usr}': "
                                 f"HTTP {r.status} {r.reason}")
            return None
        _strainer_tz = bs4.SoupStrainer(name="select", id="SystemConfiguration_timezone")
        _bs = bs4.BeautifulSoup(_raw_html, features="lxml", parse_only=_strainer_tz)
        _opts = _bs.find_all(name="option")
        for _o in _opts:
            if isinstance(_o.get("selected"), str):
                _c["tz"] = _o["value"]
        # _strainer_langs = bs4.SoupStrainer(name="select", id="SystemConfiguration_timezone")
        # bs = bs4.BeautifulSoup(await r.text(), features="lxml", parse_only=_strainer_langs)
        # _opts = bs.find_all(name="option")
        # for _o in _opts:
        #     if isinstance(_o.get("selected"), str):
        #         _c["tz"] = timezone(_o["value"])
        return _c

    async def faweb_get_timereg(self,
                                domain: str,
                                usr: str,
                                tbeg: datetime,
                                tend: datetime) -> Optional[List[Dict[str, Any]]]:
        """
        Get time registration data from the  FA web API for specified domain and time interval

        :param domain: FA domain name
        :param usr: FA user login name
        :param tbeg: date and time from which to select tasks
        :param tend: date and time till which to select tasks
        """
        _s_name = self._faweb_get_session_name(domain, usr)
        # Check cookies
        if not isinstance(getcookieval(self._s[_s_name], "XSRF-TOKEN"), str):
            raise ValueError(f"no XSRF-TOKEN: {getcookieval(self._s[_s_name], 'XSRF-TOKEN')}")
        while self._s_status[_s_name].login_mutex:
            await asyncio.sleep(1.000)
        async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/api/1.0/time_registration/data",
                                         headers={"Accept":       "application/json",
                                                  "X-CSRF-TOKEN": getcookieval(self._s[_s_name], "XSRF-TOKEN")
                                                  },
                                         params={"fields": FARequests.timeregquery()},
                                         data=FARequests.timeregform(self._fa_api_draw,
                                                                     tbeg,
                                                                     tend)) as r:
            self._fa_api_draw += 1
            if r.status == 200:
                _json = await r.json()
                return _json["data"]
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA web API time reg data request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return list()

    async def faweb_get_graphql(self,
                                domain: str,
                                usr: str,
                                opname: str,
                                pagesize: int,
                                objecttype: type,
                                query: Optional[List[Tuple[str, str]]] = None,
                                progresscallback: Optional[Callable[[int, int, int, float, str], None]] = None,
                                attempts: int = 5,
                                retry_timeout: float = 5.0,
                                retry_multiplier: float = 3.0) -> Optional[Union[Dict[str, Any], List[Any]]]:
        """
        Request generic FA web GraphQL API and return FA data output

        :param domain: FA domain name
        :param usr: FA user login name
        :param opname: name of GraphQL request
        :param pagesize: page size to use max 500
        :param objecttype: expected response object type, use list if expected collection of objects, and dict if
                           only one object
        :param query: query parameters list, each list item is query parameter tuple (tag, value) where tag is
                      tempalte tag that will be replaced by value in the text of GraphQL request, please make tags
                      unique and following naming convention applicable to programming variables, value is any text to
                      be substituted in place of tag; reserved tag names TEMPLATE_PAGE_NUM and TEMPLATE_PAGE_SIZE,
                      using of these two tag names must be avoided, because these tags are user for pagination
        :param progresscallback: callback function, accepts four parameters:
                                 cprocessed - integer how many entities processed,
                                 csuccessful - integer how many entities successfully processed,
                                 ctotal - integer how many entities in total,
                                 cprc - float in range 0.0 to 1.0 meaning % of progress
                                 cmsg - string with arbitrary message
        :param attempts: how many times to retry, one common attempts counter works for whole request for all pages
        :param retry_timeout: timeout in seconds before first retry
        :param retry_multiplier: multiplier of timeout for each consecutive retry
        :return: list of objects loaded from FA web API response JSON in case of
        """
        _current_attempt = 1
        _retry_timeout = retry_timeout
        _s_name = self._faweb_get_session_name(domain, usr)
        try:
            progresscallback(0,
                             0,
                             0,
                             0.0,
                             f"Requesting {opname} for domain {domain}")
            if objecttype == list:
                _return: Union[Dict[str, Any], List[Any]] = list()
            elif objecttype == dict:
                _return: Union[Dict[str, Any], List[Any]] = dict()
            else:
                raise ValueError(f"wrong objecttype value: {objecttype}, only list and dict supported")
            if self._s_status[_s_name].login_timestamp is None:
                raise ValueError(f"user was not authenticated before requesting")
            if not isinstance(pagesize, int) or pagesize < 1 or pagesize > 500:
                raise ValueError(f"wrong page size = {pagesize} allowed from 1 to 500")
            if opname not in self._graphql.keys():
                raise KeyError(f"source code for GraphQL request '{opname}' not found")
            _gql_query: str = self._graphql[opname].replace("TEMPLATE_PAGE_SIZE", str(pagesize))
            if query is not None:
                for _template_item in query:
                    _gql_query = _gql_query.replace(_template_item[0], _template_item[1])
            # Check cookies
            if not isinstance(getcookieval(self._s[_s_name], "XSRF-TOKEN"), str):
                raise ValueError(f"no XSRF-TOKEN: {getcookieval(self._s[_s_name], 'XSRF-TOKEN')}")
            # Request by pages
            _page_num = 1
            _records_total: Optional[int] = None
            _r_json_no_paginator = True
            while _records_total is None or len(_return) < _records_total:
                while self._s_status[_s_name].login_mutex:
                    await asyncio.sleep(1.000)
                r = await self._s[_s_name].post(f"https://{domain}.facilityapps.com/api/graphql",
                                                headers={"Accept":          "application/json",
                                                         "Origin":          f"https://{domain}.facilityapps.com",
                                                         "Accept-Encoding": "gzip, deflate",
                                                         "Accept-Language": "*",
                                                         "Content-Type":    "application/json",
                                                         "Sec-Fetch-Dest":  "empty",
                                                         "Sec-Fetch-Mode":  "cors",
                                                         "Sec-Fetch-Site":  "same-origin",
                                                         "TE":              "gzip"
                                                         },
                                                data=dumps({"operationName": opname,
                                                            "query": _gql_query.replace("TEMPLATE_PAGE_NUM",
                                                                                        str(_page_num)),
                                                            "variables": {}}, ensure_ascii=False).encode("utf-8"))
                _response_payload = await r.read()
                _response_timestamp = time()
                if _response_payload:
                    _response_content = {r.headers["Content-Type"]: _response_payload.decode("utf-8")}
                else:
                    _response_content = None
                if r.ok and _response_payload:
                    _r_json_ok = False
                    _r_json_data: Optional[List[Any]] = None
                    _r_json_paginator: Optional[Dict[str, int]] = None
                    with suppress(JSONDecodeError):
                        _r_json = loads(_response_payload.decode("utf-8"))
                        if "data" in _r_json.keys():
                            if isinstance(_r_json["data"], dict) and len(_r_json["data"]) > 0:
                                _r_json_val0 = list(_r_json["data"].values())[0]
                                if isinstance(_r_json_val0, dict):
                                    if objecttype == list and "data" in _r_json_val0.keys():
                                        if isinstance(_r_json_val0["data"], list):
                                            if "paginatorInfo" in _r_json_val0.keys():
                                                _r_json_data = _r_json_val0["data"]
                                                _r_json_pg = _r_json_val0["paginatorInfo"]
                                                _r_json_ok = True
                                                for _r_json_data_item in _r_json_data:
                                                    if isinstance(_r_json_data_item, dict):
                                                        _r_json_data_item["__response_timestamp"] = _response_timestamp
                                                _return += _r_json_data
                                                if isinstance(_r_json_pg, dict) and "total" in _r_json_pg.keys():
                                                    if isinstance(_r_json_pg["total"], int):
                                                        if _records_total is None:
                                                            _records_total = _r_json_pg["total"]
                                                            _r_json_no_paginator = False
                                                        else:
                                                            if _r_json_pg["total"] != _records_total:
                                                                raise ValueError(f"request cancelled: total number of"
                                                                                 f" records changed while paginating"
                                                                                 f" - was {_records_total}"
                                                                                 f" became {_r_json_pg['total']}")
                                    elif objecttype == dict:
                                        if len(_r_json_val0) > 0:
                                            _return = list(_r_json_val0.values())[0]
                                            if isinstance(_return, dict):
                                                _return["__response_timestamp"] = _response_timestamp
                                            _r_json_ok = True
                                    else:
                                        raise ValueError(f"wrong object in response from API"
                                                         f" for expected type {objecttype}")
                    if _r_json_ok:
                        if objecttype == dict:
                            if self._debuglog:
                                await self._base_log(i_stack(),
                                                     "debug",
                                                     (f"OK FA web API GraphQL request {opname}"
                                                      f" at page number {_page_num} / page size {pagesize}"
                                                      f" domain '{domain}' user '{usr}' {len(_return)}"
                                                      f" total records received"),
                                                     logo=_response_content)
                            progresscallback(1,
                                             1,
                                             1,
                                             1.0,
                                             f"Requesting {opname} for domain {domain}")
                            break
                        if _r_json_no_paginator:
                            if self._debuglog:
                                await self._base_log(i_stack(),
                                                     "debug",
                                                     (f"FA web API GraphQL request {opname} no paginatorInfo received"
                                                      f" at page number {_page_num} / page size {pagesize}"
                                                      f" domain '{domain}' user '{usr}' break and return "
                                                      f" with only {len(_return)} records received"),
                                                     logo=_response_content)
                            progresscallback(len(_return),
                                             len(_return),
                                             len(_return),
                                             1.0,
                                             f"Requesting {opname} for domain {domain}")
                            break
                        else:
                            if self._debuglog:
                                await self._base_log(i_stack(),
                                                     "debug",
                                                     (f"OK FA web API GraphQL request {opname}"
                                                      f" at page number {_page_num} / page size {pagesize}"
                                                      f" domain '{domain}' user '{usr}' {len(_return)}"
                                                      f" total records received"),
                                                     logo=_response_content)
                            _page_num += 1
                            _retry_timeout = retry_timeout
                            progresscallback(len(_return),
                                             len(_return),
                                             _records_total,
                                             len(_return) / _records_total if _records_total > 0 else 1.0,
                                             f"Requesting {opname} for domain {domain}")
                    else:
                        await self._base_log(i_stack(),
                                             "warning",
                                             (f"FAILED FA web API GraphQL request {opname} wrong response content"
                                              f" at page number {_page_num} / page size {pagesize}"
                                              f" domain '{domain}' user '{usr}'"),
                                             logo=_response_content)
                        return None
                else:
                    await self._base_log(i_stack(),
                                         "warning",
                                         (f"FAILED FA web API GraphQL request {opname} page number {_page_num}"
                                          f" / page size {pagesize} domain '{domain}' user '{usr}'"
                                          f" attempt {_current_attempt}: HTTP {r.status} {r.reason}"),
                                         logo=_response_content)
                    if _current_attempt < attempts:
                        _current_attempt += 1
                        await asyncio.sleep(_retry_timeout)
                        _retry_timeout *= retry_multiplier
                    else:
                        return None
            return _return
        except asyncio.exceptions.CancelledError:
            pass
        except Exception as _exc:
            await self._base_log(i_stack(),
                                 "error",
                                 (f"FAILED faweb_get_graphql() request {opname} domain '{domain}' user '{usr}':"
                                  f" {format_exc().replace(chr(10), ' ')}"),
                                 exc=_exc,
                                 fexc=format_exc())
            return None

    async def faweb_get_tasks(self,
                              domain: str,
                              usr: str,
                              siteid: int,
                              td: datetime) -> Optional[Dict[str, Any]]:
        """
        Gets all tasks from FA web API for specified domain, date and specified site ID

        :param domain: FA domain name
        :param usr: FA user login name
        :param siteid: FA backend site ID
        :param td: date for which to select tasks
        :return: dict loaded from FA web API response JSON
        """
        _s_name = self._faweb_get_session_name(domain, usr)
        try:
            # Check cookies
            if not isinstance(getcookieval(self._s[_s_name], "XSRF-TOKEN"), str):
                raise ValueError(f"no XSRF-TOKEN: {getcookieval(self._s[_s_name], 'XSRF-TOKEN')}")
            # Get all tasks
            _json_form = {"year": str(td.year),
                          "month": str(td.month),
                          "day": str(td.day),
                          "CurrentView": "day",
                          "sites[]": str(siteid),
                          "useDates": "0",
                          "includeExceptions": "true"}
            while self._s_status[_s_name].login_mutex:
                await asyncio.sleep(1.000)
            async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/api/1.0/planning/planning_data",
                                             headers={
                                                "Accept":          "application/json",
                                                "Origin":          f"https://{domain}.facilityapps.com",
                                                "Accept-Encoding": "gzip, deflate",
                                                "Accept-Language": "*",
                                                "X-CSRF-TOKEN":    getcookieval(self._s[_s_name], "XSRF-TOKEN"),
                                                "Referer":         f"https://{domain}.facilityapps.com/planning/view",
                                                "Sec-Fetch-Dest":  "empty",
                                                "Sec-Fetch-Mode":  "cors",
                                                "Sec-Fetch-Site":  "same-origin"
                                             },
                                             data=_json_form) as r:
                if r.status == 200:
                    _json_response = await r.json()
                    if "exceptions" in _json_response.keys() and _json_response["exceptions"]:
                        await self._base_log(i_stack(),
                                             "info",
                                             f"got exception field in FA API /api/1.0/planning/planning_data "
                                             f"response JSON - domain '{domain}' user '{usr}'",
                                             logo=_json_response["exceptions"])
                    return _json_response
                else:
                    await self._base_log(i_stack(), "warning",
                                         f"FAILED FA API /api/1.0/planning/planning_data request - "
                                         f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}")
                    return None
        except Exception as _exc:
            await self._base_log(i_stack(), "error", f"FAILED faweb_get_tasks(): {format_exc().replace(chr(10), ' ')}")
            return None

    async def faweb_send_push(self,
                              domain: str,
                              usr: str,
                              useridlist: List[int],
                              msg: str) -> None:
        """
        Gets all tasks for domain from FA web API for specified period

        :param domain: FA domain name
        :param usr: FA user login name
        :param useridlist: list of FA user IDs to whom to send push
        :param msg: message to send
        """
        _s_name = self._faweb_get_session_name(domain, usr)
        try:
            if not isinstance(getcookieval(self._s[_s_name], "XSRF-TOKEN"), str):
                raise ValueError(f"no XSRF-TOKEN: {getcookieval(self._s[_s_name], 'XSRF-TOKEN')}")
            _push_json_content = dumps({"message": msg,
                                        "users": useridlist,
                                        "everyone": False},
                                       ensure_ascii=False)
            while self._s_status[_s_name].login_mutex:
                await asyncio.sleep(1.000)
            r = await self._s[_s_name].post(f"https://{domain}.facilityapps.com/api/1.0/push-messaging/send",
                                            headers={"Content-Type": "application/json; charset=utf-8",
                                                     "Accept":       "application/json",
                                                     "X-CSRF-TOKEN": getcookieval(self._s[_s_name], "XSRF-TOKEN")
                                                     },
                                            data=_push_json_content)
            if r.ok:
                pass
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA web API push message send request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason} - "
                                      f"request content was: {_push_json_content}"))
        except Exception as _exc:
            await self._base_log(i_stack(), "error", f"FAILED faweb_send_push(): {format_exc().replace(chr(10), ' ')}")

    async def faweb_update_duration(self,
                                    domain: str,
                                    usr: str,
                                    taskid: int,
                                    taskseq: int,
                                    d: timedelta) -> None:
        """
        Gets all tasks for domain from FA web API for specified period

        :param domain: FA domain name
        :param usr: FA user login name
        :param taskid: FA backend task ID
        :param taskseq: FA backend task sequence number
        :param d: duration to set
        """
        _s_name = self._faweb_get_session_name(domain, usr)
        try:
            if not isinstance(getcookieval(self._s[_s_name], "XSRF-TOKEN"), str):
                raise ValueError(f"no XSRF-TOKEN: {getcookieval(self._s[_s_name], 'XSRF-TOKEN')}")
            _json = {"id":                   str(taskid),
                     "sequence_number":      str(taskseq),
                     "_token":               getcookieval(self._s[_s_name], "XSRF-TOKEN"),
                     "duration_hour":        str((d.seconds // 60) // 60),
                     "duration_minute":      str((d.seconds // 60) % 60),
                     "duration_second":      str(d.seconds % 60),
                     "editMode":             "single",
                     "excludeExceptions":    "",
                     "task_complete_emails": 0,
                     "task_canceled_emails": 0
                     }
            while self._s_status[_s_name].login_mutex:
                await asyncio.sleep(1.000)
            r = await self._s[_s_name].post(f"https://{domain}.facilityapps.com/api/1.0/planning/save",
                                            headers={"Accept":       "application/json",
                                                     "Content-Type": "application/json; charset=utf-8",
                                                     "X-CSRF-TOKEN": getcookieval(self._s[_s_name], "XSRF-TOKEN")
                                                     },
                                            data=dumps(_json, ensure_ascii=False))
            if r.ok:
                _res = await r.json()
                if "result" not in _res.keys() or _res["result"] is not True:
                    await self._base_log(i_stack(), "warning", f"FAILED FA web API update duration request - "
                                                               f"domain '{domain}' user '{usr}': API response = {_res}")
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA web API update duration request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
        except Exception as _exc:
            await self._base_log(i_stack(), "error",
                                 f"FAILED faweb_update_duration(): {format_exc().replace(chr(10), ' ')}")

    async def faweb_get_submitted_forms(self,
                                        domain: str,
                                        usr: str,
                                        tbeg: datetime,
                                        tend: datetime) -> Optional[List[Dict[str, Any]]]:
        """
        Gets all submitted forms for domain from FA web API for specified period

        :param domain: FA domain name
        :param usr: FA user login name
        :param tbeg: date and time from which to select tasks
        :param tend: date and time till which to select tasks
        :return: dict loaded from FA web API response JSON
        """
        _s_name = self._faweb_get_session_name(domain, usr)
        try:
            # Check cookies
            if not isinstance(getcookieval(self._s[_s_name], "XSRF-TOKEN"), str):
                raise ValueError(f"no XSRF-TOKEN: {getcookieval(self._s[_s_name], 'XSRF-TOKEN')}")
            # Get all submitted forms
            while self._s_status[_s_name].login_mutex:
                await asyncio.sleep(1.000)
            r = await self._s[_s_name].post((f"https://{domain}.facilityapps.com"
                                             f"/api/1.0/forms/reporting?fields={FARequests.formsreportingquery()}"),
                                            headers={"Accept":       "application/json",
                                                     "X-CSRF-TOKEN": getcookieval(self._s[_s_name], "XSRF-TOKEN")
                                                     },
                                            data=FARequests.formsreportingform(1, tbeg, tend))
            if r.ok:
                _forms_list_json = await r.json()
                return _forms_list_json["data"]
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA web API submitted forms list request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return None
        except Exception as _exc:
            await self._base_log(i_stack(), "error", f"FAILED faweb_get_submitted_forms(): "
                                                     f"{format_exc().replace(chr(10), ' ')}")
            return None

    async def faweb_get_form_details(self,
                                     domain: str,
                                     usr: str,
                                     submissionid: int) -> Optional[Dict[str, Any]]:
        """
        Get submitted form details and parse with parse_form_html()

        :param domain: FA domain name
        :param usr: FA user login name
        :param submissionid: ID of form submission
        :return: dictionary with submitted form contents
        """
        # Output dict
        _out = dict()
        _s_name = self._faweb_get_session_name(domain, usr)
        while self._s_status[_s_name].login_mutex:
            await asyncio.sleep(1.000)
        r = await self._s[_s_name].get(f"https://{domain}.facilityapps.com/forms/submission/view/{submissionid}",
                                       headers={"Accept": "text/html, application/xhtml+xml"})
        if r.ok:
            _raw_html = await r.text()
        else:
            await self._base_log(i_stack(), "error",
                                 f"FAILED requesting details for form submission id = {submissionid} for "
                                 f"domain '{domain}' user '{usr}': "
                                 f"HTTP {r.status} {r.reason}")
            return None
        return self.parse_form_html(_raw_html)

    async def faapp_login(self, domain: str, usr: str, pwd: str) -> Optional[Dict[str, Any]]:
        """
        Login to FA mobile app API and return the main user record

        :param domain: FA domain name
        :param usr: FA user login name
        :param pwd: FA password
        :return: dict with FA user full record
        """
        _form = {"T":           "LogIn_v2",
                 "username":    usr,
                 "password":    pwd,
                 "version":     self._vapp,
                 "mode":        "LogIn",
                 "auth_method": "basic"
                 }
        _s_name = self._faapp_get_session_name(domain, usr)
        async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/FacilityAppsAPI.php",
                                         headers=self._app_headers,
                                         data=_form) as r:
            if r.status == 200:
                _json = await r.json()
                return _json
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA mobile API login request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return None

    async def faapp_get_forms(self, domain: str, usr: str, language: str, session_id: str) -> Optional[Dict[str, Any]]:
        """
        Call FA mobile API GetChecklists and return result

        :param domain: FA domain name
        :param usr: FA user login name
        :param language: language code in form of en_EN to use for language and region form fields
        :param session_id: FA mobile API session_id
        :return: dict with FA API output of GetChecklists
        """
        _form = {"session_id":  session_id,
                 "T":           "GetChecklists",
                 "L":           language,
                 "region_code": language
                 }
        _s_name = self._faapp_get_session_name(domain, usr)
        async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/FacilityAppsAPI.php",
                                         headers=self._app_headers,
                                         data=_form) as r:
            if r.status == 200:
                _json = await r.json()
                return _json
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA API GetChecklists request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return None

    async def faapp_task_begin(self,
                               domain: str,
                               usr: str,
                               language: str,
                               session_id: str,
                               siteid: int,
                               taskid: int,
                               taskseq: int,
                               stime: datetime) -> bool:
        """
        Set task to begin

        :param domain: FA domain name
        :param usr: FA user login name
        :param language: language code in form of en_EN to use for language and region form fields
        :param session_id: FA mobile API session_id
        :param siteid: FA backend site ID
        :param taskid: FA backend task ID
        :param taskseq: FA backend task sequence number
        :param stime: date and time to set status at
        :return: True if successful
        """
        _form = {"T":                "StartProgressOnTask",
                 "L":                language,
                 "FormData":         dumps({"task_id":            str(taskid),
                                            "sequence_number":    taskseq,
                                            "time_of_change":     str(int(stime.timestamp())),
                                            "latitude":           None,
                                            "longitude":          None,
                                            "time_registrations": None
                                            }, ensure_ascii=False),
                 "ReferenceId":      siteid,
                 "ReferenceType":    1,
                 "Timestamp":        int(stime.timestamp()),
                 "session_id":       session_id,
                 "AppVersionNumber": self._vapp,
                 "TaskData":         "null"
                 }
        _s_name = self._faapp_get_session_name(domain, usr)
        async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/FacilityAppsAPI.php",
                                         headers=self._app_headers,
                                         data=_form) as r:
            if r.status == 200:
                _json = await r.json()
                return _json
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA API StartProgressOnTask request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return False

    async def faapp_task_set_status(self,
                                    domain: str,
                                    usr: str,
                                    session_id: str,
                                    taskid: int,
                                    taskseq: int,
                                    status: int,
                                    stime: datetime) -> bool:
        """
        Set status to task

        :param domain: FA domain name
        :param usr: FA user login name
        :param session_id: FA mobile API session_id
        :param taskid: FA backend task ID
        :param taskseq: FA backend task sequence number
        :param status: integer FA backend status to set
        :param stime: date and time to set status at
        :return: True if successful
        """
        _form = {"T":               "SetCalendarSubStatus",
                 "task_id":         taskid,
                 "sequence_number": taskseq,
                 "option":          status,
                 "latitude":        str(),
                 "longitude":       str(),
                 "timestamp":       int(stime.timestamp()),
                 "extra":           '{"clock_log_list":null}',
                 "session_id":      session_id
                 }
        _s_name = self._faapp_get_session_name(domain, usr)
        async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/FacilityAppsAPI.php",
                                         headers=self._app_headers,
                                         data=_form) as r:
            if r.status == 200:
                _json = await r.json()
                return _json
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA API SetCalendarSubStatus request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return False

    async def faapp_task_chat(self,
                              domain: str,
                              usr: str,
                              language: str,
                              session_id: str,
                              siteid: int,
                              taskid: int,
                              taskseq: int,
                              msg: str,
                              fauserid: int,
                              dtz: tzinfo) -> bool:
        """
        Send chat message to the task

        :param domain: FA domain name
        :param usr: FA user login name
        :param language: language code in form of en_EN to use for language and region form fields
        :param session_id: FA mobile API session_id
        :param siteid: FA backend site ID
        :param taskid: FA backend task ID
        :param taskseq: FA backend task sequence number
        :param msg: text message to append to chat
        :param fauserid: FA user id under which to send message
        :param dtz: tzinfo time zone of the domain
        :return: True if successful
        """
        _time = int(time())
        _time_tz = datetime.fromtimestamp(_time, dtz)
        _form = {"T":                "SaveTaskMessage",
                 "L":                language,
                 "FormData":         dumps({"id":              0,
                                            "task_id":         taskid,
                                            "subtask_id":      None,
                                            "sequence_number": taskseq,
                                            "message":         msg,
                                            "user":            str(fauserid),
                                            "timestamp":       _time_tz.strftime("%Y-%m-%d %H:%M:%S"),
                                            "images":          []
                                            }, ensure_ascii=False),
                 "ReferenceId":      siteid,
                 "ReferenceType":    1,
                 "Timestamp":        _time,
                 "session_id":       session_id,
                 "AppVersionNumber": self._vapp,
                 "TaskData":         "null"
                 }
        _s_name = self._faapp_get_session_name(domain, usr)
        async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/FacilityAppsAPI.php",
                                         headers=self._app_headers,
                                         data=_form) as r:
            if r.status == 200:
                _json = await r.json()
                return _json
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA API StartProgressOnTask request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return False

    async def faapp_task_drop_begin(self,
                                    domain: str,
                                    usr: str,
                                    language: str,
                                    session_id: str,
                                    siteid: int,
                                    taskid: int,
                                    taskseq: int,
                                    stime: datetime) -> bool:
        """
        Set task back to blue open state

        :param domain: FA domain name
        :param usr: FA user login name
        :param language: language code in form of en_EN to use for language and region form fields
        :param session_id: FA mobile API session_id
        :param siteid: FA backend site ID
        :param taskid: FA backend task ID
        :param taskseq: FA backend task sequence number
        :param stime: date and time to set status at
        :return: True if successful
        """
        _form = {"T":                "TaskToBacklog",
                 "L":                language,
                 "FormData":         dumps({"task_id":            str(taskid),
                                            "sequence_number":    taskseq,
                                            "time_of_change":     str(int(stime.timestamp())),
                                            "latitude":           None,
                                            "longitude":          None,
                                            "time_registrations": list()
                                            }, ensure_ascii=False),
                 "ReferenceId":      siteid,
                 "ReferenceType":    1,
                 "Timestamp":        int(stime.timestamp()),
                 "session_id":       session_id,
                 "AppVersionNumber": self._vapp,
                 "TaskData":         "null"
                 }
        _s_name = self._faapp_get_session_name(domain, usr)
        async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/FacilityAppsAPI.php",
                                         headers=self._app_headers,
                                         data=_form) as r:
            if r.status == 200:
                _json = await r.json()
                return _json
            else:
                await self._base_log(i_stack(),
                                     "warning",
                                     (f"FAILED FA API TaskToBacklog request - "
                                      f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}"))
                return False

    async def faapp_get_tasks(self,
                              domain: str,
                              usr: str,
                              session_id: str,
                              tbeg: datetime,
                              tend: datetime,
                              language: str,
                              sitelist: List[int]) -> Optional[Dict[str, Any]]:
        """
        Gets all tasks for domain from FA mobile API for specified period

        :param domain: FA domain name
        :param usr: FA user login name
        :param session_id: FA mobile API session_id
        :param tbeg: date and time from which to select tasks
        :param tend: date and time till which to select tasks
        :param language: language code in form of en_EN to use for language and region form fields
        :param sitelist: list of integer FA backend site IDs
        :return: dict loaded from FA web API response JSON
        """
        _s_name = self._faapp_get_session_name(domain, usr)
        try:
            _json = {"CurrentView":       "custom",
                     "start":             tbeg.strftime("%Y-%m-%d"),
                     "end":               tend.strftime("%Y-%m-%d"),
                     "sites":             sitelist,
                     "includeExceptions": True,
                     "region_code":       language
                     }
            async with self._s[_s_name].post(f"https://{domain}.facilityapps.com/api/1.0/planning/planning_data",
                                             headers={**self._app_headers,
                                                      **{"Content-Type":  "application/json; charset=utf-8",
                                                         "Authorization": f"Bearer {session_id}"
                                                         }
                                                      },
                                             data=dumps(_json, ensure_ascii=False)) as r:
                if r.status == 200:
                    _json_response = await r.json()
                    if "exceptions" in _json_response.keys() and _json_response["exceptions"]:
                        await self._base_log(i_stack(),
                                             "info",
                                             f"got exception field in FA API /api/1.0/planning/planning_data "
                                             f"response JSON - domain '{domain}' user '{usr}'",
                                             logo=_json_response["exceptions"])
                    return _json_response
                else:
                    await self._base_log(i_stack(), "warning",
                                         f"FAILED FA API /api/1.0/planning/planning_data request - "
                                         f"domain '{domain}' user '{usr}': HTTP {r.status} {r.reason}")
                    return None
        except Exception as _exc:
            await self._base_log(i_stack(), "error", str(), _exc, format_exc())
            return None

    def _faapp_get_session_name(self, domain: str, usr: str) -> str:
        """
        Generate aiohttp session name and create the session itself for FA mobile API to persist it

        :param domain: FA domain name
        :param usr: FA user login name
        :return: generated aiohttp session name to persiste the session in self._s
        """
        _s_name = f"app-{domain}-{usr}".lower()
        if _s_name not in self._s.keys():
            self._s[_s_name] = aiohttp.ClientSession()
            self._s_status[_s_name] = HTTPClientSessionStatus(None, None)
        return _s_name

    def _faweb_get_session_name(self, domain: str, usr: str) -> str:
        """
        Generate aiohttp session name and create the session itself for FA web API to persist it

        :param domain: FA domain name
        :param usr: FA user login name
        :return: generated aiohttp session name to persiste the session in self._s
        """
        _s_name = f"web-{domain}-{usr}".lower()
        if _s_name not in self._s.keys():
            self._s[_s_name] = aiohttp.ClientSession()
        if _s_name not in self._s_status.keys():
            self._s_status[_s_name] = HTTPClientSessionStatus(None, None)
        return _s_name



# coding=utf-8
"""
Base class for the service
"""
import asyncio
import hashlib
import time
from copy import deepcopy
from datetime import datetime
from inspect import FrameInfo, stack as i_stack
from json import dumps, loads
from os import walk
from os.path import splitext
from traceback import format_exc
from typing import Any, Callable, Coroutine, Dict, List, Optional, Union

import aio_pika
import asyncpg.prepared_stmt

# import sys
# sys.path.insert(1, "/home/deviqnfc/dev/samgenericservices")

from .BaseMicroService import BaseMicroService
from .FAAPIService import FAAPIService

# from BaseMicroService import BaseMicroService   # nopep8
# from FAAPIService import FAAPIService           # nopep8


class IQNFCService(BaseMicroService):
    """
    Base class for the service
    """

    def __init__(self,
                 evloop: asyncio.AbstractEventLoop,
                 appname: str,
                 configfile: str,
                 dbname: Optional[str] = None,
                 servicename: Optional[str] = None,
                 tcphost: Optional[str] = None,
                 tcpport: Optional[int] = None,
                 httpmwares: Optional[List] = None,
                 pikattl: int = 60):
        """
        Init the base class

        :param evloop: asyncio event loop to run in
        :param appname: name of the application
        :param configfile: config file name
        :param dbname: name of database to use
        :param servicename: name of the current service
        :param tcphost: optional TCP host address for web server to listen at
        :param tcpport: optional TCP port number for web server to listen at
        :param httpmwares: list of aiohttp middleware callables as is passed to aiohttp.web.Application
                           middlewares= option
        :param pikattl: RabbitMQ messages ttl
        """
        super().__init__(evloop, appname, configfile, dbname, servicename, tcphost, tcpport, httpmwares, pikattl)
        # SQL statements
        self._q = {
            "get_domains":            "SELECT * FROM public.fadomains;",

            "get_domain":             "SELECT * FROM public.fadomains WHERE fa_domain = $1;",

            "check_domain":           "SELECT enabled FROM public.fadomains WHERE fa_domain = $1;",

            "get_admin_record":       "SELECT * FROM public.adminusers WHERE fa_domain = $1;",

            "add_admin":              "INSERT INTO public.adminusers VALUES ($1, $2, $3, $4, $5, $6) "
                                      "RETURNING fa_domain;",

            "upd_admin_session":      "UPDATE public.adminusers "
                                      "SET fa_session_id = $2, fa_session_id_timestamp = $3 "
                                      "WHERE fa_domain = $1 "
                                      "RETURNING fa_session_id;",

            "upsert_fa_user":         "INSERT INTO public.refusers VALUES ($1, $2, $3, $4, $5, "
                                      "$6, $7, $8, $9, $10, $11, $12) "
                                      "ON CONFLICT (fa_domain, fa_id) DO UPDATE SET "
                                      "(fa_username, fa_name, fa_email, fa_phone, "
                                      "fa_disable_login, fa_created_at, fa_updated_at, sync_timestamp, "
                                      "search_string, search_list)"
                                      " = "
                                      "($3, $4, $5, $6, $7, $8, $9, $10, $11, $12) "
                                      "RETURNING fa_domain, fa_id;",

            "upsert_fa_site":         "INSERT INTO public.refsites VALUES ($1, $2, $3, $4, $5, "
                                      "$6, $7, $8, $9, $10, $11, $12) "
                                      "ON CONFLICT (fa_domain, fa_id) DO UPDATE SET "
                                      "(fa_key, fa_name, fa_region_name, fa_address_city, fa_address, fa_created_at, "
                                      "fa_updated_at, sync_timestamp, search_string, search_list)"
                                      " = "
                                      "($3, $4, $5, $6, $7, $8, $9, $10, $11, $12) "
                                      "RETURNING fa_domain, fa_id;",

            "upsert_fa_floor":        "INSERT INTO public.reffloors VALUES ($1, $2, $3, $4, $5, "
                                      "$6, $7, $8, $9, $10, $11) "
                                      "ON CONFLICT (fa_domain, fa_site_id, fa_floor_id) DO UPDATE SET "
                                      "(fa_floor_name, fa_floor_type, fa_floor_level, "
                                      "fa_created_at, fa_updated_at, sync_timestamp, search_string, search_list)"
                                      " = "
                                      "($4, $5, $6, $7, $8, $9, $10, $11) "
                                      "RETURNING fa_domain, fa_site_id, fa_floor_id;",

            "upsert_fa_space":        "INSERT INTO public.refspaces VALUES ($1, $2, $3, $4, $5, "
                                      "$6, $7, $8, $9, $10, $11, $12) "
                                      "ON CONFLICT (fa_domain, fa_site_id, fa_floor_id, fa_space_id) "
                                      "DO UPDATE SET "
                                      "(fa_space_nr, fa_space_order, fa_space_name, fa_created_at, "
                                      "fa_updated_at, sync_timestamp, search_string, search_list)"
                                      " = "
                                      "($5, $6, $7, $8, $9, $10, $11, $12) "
                                      "RETURNING fa_domain, fa_site_id, fa_floor_id, fa_space_id;",

            "search_site":            "SELECT * FROM public.refsites WHERE fa_domain = $1 AND fa_name = $2;",

            "search_site_by_key":     "SELECT * FROM public.refsites WHERE fa_domain = $1 AND fa_key = $2;",

            "search_site_loose":      "SELECT * FROM public.refsites WHERE fa_domain = $1 AND search_string = $2 "
                                      "ORDER BY "
                                      "CASE WHEN sync_timestamp IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE sync_timestamp END DESC, "
                                      "CASE WHEN fa_created_at IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE fa_created_at END DESC;",

            "get_site":               "SELECT * FROM public.refsites WHERE fa_domain = $1 AND fa_id = $2;",

            "get_sites":              "SELECT * FROM public.refsites WHERE fa_domain = $1;",

            "get_sites_monitored":    "SELECT _s.* FROM"
                                      " public.refsites AS _s,"
                                      " public.fadomains AS _d,"
                                      " unnest(_d.sites_tasks_monitoring) AS _sid(fa_id) "
                                      "WHERE _d.fa_domain = $1 AND _s.fa_domain = _d.fa_domain"
                                      " AND _s.fa_id = _sid.fa_id;",

            "search_floor":           "SELECT * FROM public.reffloors "
                                      "WHERE fa_domain = $1 AND fa_floor_name = $2 AND fa_site_id = $3 "
                                      "ORDER BY "
                                      "CASE WHEN sync_timestamp IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE sync_timestamp END DESC, "
                                      "CASE WHEN fa_created_at IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE fa_created_at END DESC;",

            "search_floor_loose":     "SELECT * FROM public.reffloors WHERE fa_domain = $1 AND search_string = $2 "
                                      "ORDER BY "
                                      "CASE WHEN sync_timestamp IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE sync_timestamp END DESC, "
                                      "CASE WHEN fa_created_at IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE fa_created_at END DESC;",

            "get_floor":              "SELECT * FROM public.reffloors "
                                      "WHERE fa_domain = $1 AND fa_site_id = $2 AND fa_floor_id = $3;",

            "search_space":           "SELECT * FROM public.refspaces "
                                      "WHERE fa_domain = $1 AND fa_space_name = $2 "
                                      "AND fa_site_id = $3 AND fa_floor_id = $4 "
                                      "ORDER BY "
                                      "CASE WHEN sync_timestamp IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE sync_timestamp END DESC, "
                                      "CASE WHEN fa_created_at IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE fa_created_at END DESC;",

            "search_space_loose":     "SELECT * FROM public.refspaces WHERE fa_domain = $1 AND search_string = $2 "
                                      "ORDER BY "
                                      "CASE WHEN sync_timestamp IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE sync_timestamp END DESC, "
                                      "CASE WHEN fa_created_at IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE fa_created_at END DESC;",

            "get_space":              "SELECT * FROM public.refspaces "
                                      "WHERE fa_domain = $1 AND fa_site_id = $2 "
                                      "AND fa_floor_id = $3 AND fa_space_id = $4;",

            "search_user":            "SELECT * FROM public.refusers WHERE fa_domain = $1 AND fa_name = $2 "
                                      "ORDER BY "
                                      "CASE WHEN sync_timestamp IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE sync_timestamp END DESC, "
                                      "CASE WHEN fa_created_at IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE fa_created_at END DESC;",

            "search_user_loose":      "SELECT * FROM public.refusers WHERE fa_domain = $1 AND search_string = $2 "
                                      "ORDER BY "
                                      "CASE WHEN sync_timestamp IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE sync_timestamp END DESC, "
                                      "CASE WHEN fa_created_at IS NULL THEN"
                                      " make_timestamp(1970, 01, 01, 00, 00, 00.000) ELSE fa_created_at END DESC;",

            "get_user":               "SELECT * FROM public.refusers WHERE fa_domain = $1 AND fa_id = $2;",

            "get_users":              "SELECT * FROM public.refusers WHERE fa_domain = $1;",

            "update_domain":          "UPDATE public.fadomains SET tz = $2 WHERE fa_domain = $1;",

            "insert_log":             "INSERT INTO public.log "
                                      "VALUES (nextval('idcounter'), "
                                      "now(), $1, $2, $3, $4::jsonb, $5, $6, $7);",

            "get_logs_by_key":        "SELECT * FROM public.log WHERE key = $1 "
                                      "AND timestamp > (now() - make_interval(secs := $2));",

            "insert_timereg":         "INSERT INTO public.timeregs VALUES ($1, $2, $3, $4, $5, "
                                      "$6, $7, $8, $9, $10) RETURNING *;",

            "get_timeregs":           "SELECT * FROM public.timeregs WHERE fa_domain = $1 AND fa_time_tz >= $2;",

            "get_timereg_keys":       "SELECT key FROM public.timeregs "
                                      "WHERE timeregs.fa_domain = $1 AND fa_time_tz >= $2;",

            "get_tasks":              "SELECT * FROM public.reftasks WHERE "
                                      "fa_domain = $1 AND fa_time_start >= $2;",

            "get_task":               "SELECT * FROM public.reftasks WHERE "
                                      "fa_domain = $1 AND fa_task_id = $2 AND fa_task_sequence = $3;",

            "upsert_task":            "INSERT INTO public.reftasks "
                                      "VALUES ($1, $2, $3, $4::jsonb, $5, $6, $7, "
                                      "$8, $9, $10, $11, $12, $13, $14, $15, $16) "
                                      "ON CONFLICT (fa_domain, fa_task_id, fa_task_sequence) DO UPDATE SET "
                                      "(json, fa_site_id, fa_floor_id, fa_space_id, fa_owner_id, "
                                      "fa_time_start, fa_time_end, nfc_status, fa_substatus, fa_status, "
                                      "fa_translation, nfc_algorithm, nfc_status_timestamp) "
                                      " = "
                                      "($4::jsonb, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16)"
                                      "RETURNING fa_domain, fa_task_id, fa_task_sequence;",

            "update_task_nfc_status": "UPDATE public.reftasks SET nfc_status = $4, "
                                      "nfc_status_timestamp = $5 WHERE "
                                      "fa_domain = $1 AND fa_task_id = $2 AND fa_task_sequence = $3 "
                                      "RETURNING nfc_status;",

            "upsert_nfc_ref":         "INSERT INTO public.tasknfcs "
                                      "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10) "
                                      "ON CONFLICT (fa_domain, fa_site_id, fa_floor_id, "
                                      "fa_space_id, task_order, nfc_action) "
                                      "DO UPDATE SET "
                                      "(fa_task_id, fa_task_sequence, "
                                      "nfc_algorithm, nfc_index) = ($7, $8, $9, $10) "
                                      "RETURNING fa_task_id, fa_task_sequence, nfc_algorithm, nfc_index;",

            "get_nfc_ref":  "SELECT tnfc.task_order, tnfc.fa_task_id, tnfc.fa_task_sequence, tnfc.nfc_action, "
                            "tsk.fa_time_start, tsk.fa_time_end, tsk.fa_owner_id, tsk.nfc_status, "
                            "tsk.fa_substatus, tsk.fa_status, tsk.json, tnfc.nfc_algorithm, "
                            "tnfc.nfc_index, tsk.nfc_status_timestamp "
                            "FROM public.tasknfcs AS tnfc, public.reftasks AS tsk "
                            "WHERE tnfc.fa_domain = tsk.fa_domain "
                            "AND tnfc.fa_task_id = tsk.fa_task_id AND tnfc.fa_task_sequence = tsk.fa_task_sequence "
                            "AND tnfc.fa_domain = $1 AND tnfc.fa_site_id = $2 "
                            "AND tnfc.fa_floor_id = $3 AND tnfc.fa_space_id = $4;",

            "delete_nfc_ref":         "DELETE FROM public.tasknfcs WHERE fa_domain = $1;",

            "get_overdue":  "SELECT t.fa_domain, t.fa_task_id, t.fa_task_sequence, t.fa_site_id, s.fa_name, "
                            "s.fa_region_name, t.fa_translation, t.fa_floor_id, t.fa_space_id, t.fa_substatus,"
                            "t.fa_status, t.fa_time_start, t.fa_time_end, t.fa_owner_id, t.nfc_algorithm "
                            "FROM public.reftasks AS t, public.refsites AS s "
                            "WHERE s.fa_domain = t.fa_domain AND s.fa_id = t.fa_site_id AND t.fa_domain = $1 "
                            "AND (t.fa_status NOT IN (2, 3, 4) OR t.fa_status IS NULL) AND t.nfc_status NOT IN (2, 3) "
                            "AND t.fa_time_end >= $2 AND t.fa_time_end < $3;",

            "get_not_begun":    "SELECT t.fa_domain, t.fa_task_id, t.fa_task_sequence, t.fa_site_id, s.fa_name, "
                                "s.fa_region_name, t.fa_translation, t.fa_floor_id, t.fa_space_id, t.fa_substatus,"
                                "t.fa_status, t.fa_time_start, t.fa_time_end, t.fa_owner_id, t.nfc_algorithm "
                                "FROM public.reftasks AS t, public.refsites AS s "
                                "WHERE s.fa_domain = t.fa_domain AND s.fa_id = t.fa_site_id AND t.fa_domain = $1 "
                                "AND t.nfc_status IN (-1, 0) "
                                "AND t.fa_time_start >= $2 AND t.fa_time_start < $3;",

            "get_manual_begun":     "SELECT t.fa_domain, t.fa_task_id, t.fa_task_sequence, t.fa_site_id, s.fa_name, "
                                    "s.fa_region_name, t.fa_translation, t.fa_floor_id, t.fa_space_id, t.fa_substatus,"
                                    "t.fa_status, t.fa_time_start, t.fa_time_end, t.fa_owner_id, t.nfc_algorithm "
                                    "FROM public.reftasks AS t, public.refsites AS s "
                                    "WHERE s.fa_domain = t.fa_domain AND s.fa_id = t.fa_site_id AND t.fa_domain = $1 "
                                    "AND t.fa_status = 1 AND t.fa_substatus IS NULL AND t.nfc_status = -1 "
                                    "AND t.fa_time_start >= $2;",

            "upsert_submitted_form":  "INSERT INTO public.formssubmitted "
                                      "VALUES ($1, $2, $3, $4, $5, $6, $7, $8::jsonb) "
                                      "ON CONFLICT (fa_domain, fa_form_id, fa_submision_id, fa_timestamp) "
                                      "DO UPDATE SET "
                                      "(fa_site_id, fa_element_id, fa_user_id, parsed_json) "
                                      "= ($5, $6, $7, $8::jsonb) "
                                      "RETURNING fa_domain, fa_form_id, fa_submision_id, fa_timestamp;",

            "get_submitted_forms":    "SELECT * FROM public.formssubmitted WHERE fa_domain = $1 "
                                      "AND fa_form_id = $2 AND fa_timestamp >= $3 AND fa_timestamp < $4;",

            "get_site_timeregs":      "SELECT * FROM public.timeregs "
                                      "WHERE fa_domain = $1 AND fa_time_tz >= $2 AND fa_time_tz < $3 "
                                      "AND NOT space_precise ORDER BY fa_time_tz;",

            "get_space_timeregs":     "SELECT fa_site_id, fa_user_id, COUNT(key) AS fa_user_id_cnt "
                                      "FROM public.timeregs WHERE fa_domain = $1 "
                                      "AND space_precise AND fa_time_tz >= $2 AND fa_time_tz < $3 "
                                      "GROUP BY fa_site_id, fa_user_id;",

            "upsert_siteregreport":   "INSERT INTO public.siteregreport "
                                      "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9) "
                                      "ON CONFLICT (fa_domain, fa_site_id, fa_user_id, fa_date) "
                                      "DO UPDATE SET "
                                      "(fa_time_tz_beg, fa_time_tz_end, "
                                      "time_interval, space_checkins, site_checkins) = ($5, $6, $7, $8, $9) "
                                      "RETURNING fa_domain, fa_site_id, fa_user_id, fa_date;"

        }
        # FA domains and admin users list to serve
        self._domains: List[asyncpg.Record] = list()
        self._admin_records = list()
        # Load GraphQL requests
        if "graphqlpath" in self._config.keys():
            _graphql_ref = dict()
            for _tmp_root, _tmp_dirs, _tmp_files in walk(self._config["graphqlpath"]):
                for _tmp_name_walk_f in _tmp_files:
                    _graphql_filename = _tmp_root + "/" + _tmp_name_walk_f
                    with open(_graphql_filename, "rb") as _graphql_fd:
                        _graphql_ref[splitext(_tmp_name_walk_f)[0]] = _graphql_fd.read().decode("utf-8")
        else:
            _graphql_ref = None
        # Create the FA API service
        self._fa = FAAPIService(self._log_cloud,
                                self._config["vapp"],
                                self._config["uagent"],
                                self._config["intervals"]["faweb_relogin"],
                                _graphql_ref,
                                True if self._config["loglevel"].upper() == "DEBUG" else False)
        # Container for logged in domains / users
        self._admin_records_auth = list()
        # Standard queues to listen
        self._queue_control: Optional[aio_pika.queue.Queue] = None
        # Control commands handlers
        # key is command name,
        # value is callable coroutine, it should accept only one argument - json received from RabbitMQ control queue
        self._control_handlers: Dict[str, Callable[..., Coroutine[Any, Any, Any]]] = dict()

    async def declare_standard_queues(self) -> None:
        """
        Declaring standard queues - control and report
        """
        pass
        # Queue for handling control messages
        self._queue_control = await self._rmq_rcv_channel.declare_queue(f"{self._dbname}.{self._servicename}.control",
                                                                        durable=True)
        await self._queue_control.consume(self._rmq_handle_control)
        # # Queue for handling responses from other services in reply to requests sent from this service
        # await self._rmq_snd_channel.declare_queue(f"{self._dbname}.{self._servicename}.get.responses",
        #                                           durable=True)

    async def get_admin_records(self) -> None:
        """
        Get domains list and all enabled domain records
        """
        async with self._pg.acquire() as _c:
            self._domains = await _c.fetch(self._q["get_domains"])
        _admin_records = list()
        for _domain in self._domains:
            if _domain["enabled"]:
                async with self._pg.acquire() as _c:
                    _admin_record = await _c.fetch(self._q["get_admin_record"],
                                                   _domain["fa_domain"])
                if len(_admin_record) == 1:
                    _admin_record_dict = dict(_admin_record[0])
                    _admin_record_dict["tz"] = _domain["tz"]
                    _admin_records.append(_admin_record_dict)
                    await self._log_cloud(i_stack(),
                                          "debug",
                                          (f"OK FA domain '{_admin_record[0]['fa_domain']}' "
                                           f"found admin user '{_admin_record[0]['fa_usr']}' "
                                           f"language {_admin_record[0]['fa_forms_language']}"))
                else:
                    await self._log_cloud(i_stack(),
                                          "debug",
                                          (f"found {len(_admin_record)} admin records "
                                           f"for domain '{_domain['fa_domain']}' but must be only single one, "
                                           f"skipping domain"))
        self._admin_records = deepcopy(_admin_records)

    async def login_to_fa(self) -> None:
        """
        Login to FA web API and create corresponding sessions
        """
        # First get admin records to serve
        await self.get_admin_records()
        # Login to all FA web API records
        _admin_records_auth = list()
        for _admin_record in self._admin_records:
            _fa_web_login_res = await self._fa.faweb_login(_admin_record["fa_domain"],
                                                           _admin_record["fa_usr"],
                                                           _admin_record["fa_pwd"])
            if _fa_web_login_res:
                _admin_records_auth.append(_admin_record)
                await self._log_cloud(i_stack(),
                                      "info",
                                      (f"OK logged in to domain '{_admin_record['fa_domain']}' "
                                       f"user '{_admin_record['fa_usr']}'"))
            else:
                await self._log_cloud(i_stack(),
                                      "warning",
                                      (f"can not login to domain '{_admin_record['fa_domain']}' "
                                       f"user '{_admin_record['fa_usr']}'"))
        self._admin_records_auth = deepcopy(_admin_records_auth)

    async def logout_from_fa(self) -> None:
        """
        Log out accounts from FA API
        """
        for _admin_record_auth in self._admin_records_auth:
            _res = await self._fa.faweb_logout(_admin_record_auth["fa_domain"],
                                               _admin_record_auth["fa_usr"])
            if _res:
                self._log.info(f"OK web logout successful for user '{_admin_record_auth['fa_usr']}' "
                               f"from domain '{_admin_record_auth['fa_domain']}'")
            else:
                self._log.warning(f"FAILED web logout for user '{_admin_record_auth['fa_usr']}' "
                                  f"from domain '{_admin_record_auth['fa_domain']}'")
        self._admin_records_auth = list()

    async def search_site(self, domain: str, name: str) -> Optional[int]:
        """
        Searches site by name, there should be only one, otherwise returns first match and logs error

        :param domain: FA domain name
        :param name: name to search for
        :return: ID of entry in FA backend
        """
        try:
            async with self._pg.acquire() as _c:
                _query_r = await _c.fetch(self._q["search_site"], domain, name)
            if len(_query_r) == 1:
                return _query_r[0]["fa_id"]
            elif len(_query_r) == 0:
                return None
            else:
                raise ValueError(f"{len(_query_r)} entries found")
        except Exception as _exc:
            await self._log_cloud(i_stack(), "error", str(), _exc, format_exc())
            return None

    async def search_site_by_key(self, domain: str, key: str) -> Optional[int]:
        """
        Searches site by key id string, there should be only one, otherwise returns first match and logs error

        :param domain: FA domain name
        :param key: name to search for
        :return: ID of entry in FA backend
        """
        try:
            async with self._pg.acquire() as _c:
                _query_r = await _c.fetch(self._q["search_site_by_key"], domain, key)
            if len(_query_r) == 1:
                return _query_r[0]["fa_id"]
            elif len(_query_r) == 0:
                return None
            else:
                raise ValueError(f"{len(_query_r)} entries found")
        except Exception as _exc:
            await self._log_cloud(i_stack(), "error", str(), _exc, format_exc())
            return None

    async def search_floor(self, domain: str, name: str, siteid: int) -> Optional[int]:
        """
        Searches floor by name, there should be only one, otherwise returns first match and logs error

        :param domain: FA domain name
        :param name: name to search for
        :param siteid: FA backed ID of site
        :return: ID of entry in FA backend
        """
        try:
            async with self._pg.acquire() as _c:
                _query_r = await _c.fetch(self._q["search_floor"], domain, name, siteid)
            if len(_query_r) == 1:
                return _query_r[0]["fa_floor_id"]
            elif len(_query_r) == 0:
                return None
            else:
                await self._log_cloud(i_stack(),
                                      "warning",
                                      f"search_floor(): {len(_query_r)} entries found "
                                      f"with name '{name}' site ID = {siteid} "
                                      f"please check database for outdated records")
                return _query_r[0]["fa_floor_id"]
        except Exception as _exc:
            await self._log_cloud(i_stack(), "error", str(), _exc, format_exc())
            return None

    async def search_space(self,
                           domain: str,
                           name: str,
                           siteid: int,
                           floorid: Optional[int] = None) -> Optional[int]:
        """
        Searches space by name, there should be only one, otherwise returns first match and logs error

        :param domain: FA domain name
        :param name: name to search for
        :param siteid: FA backed ID of site
        :param floorid: FA backed ID of floor
        :return: ID of entry in FA backend
        """
        try:
            if floorid is None:
                return None
            async with self._pg.acquire() as _c:
                _query_r = await _c.fetch(self._q["search_space"], domain, name, siteid, floorid)
            if len(_query_r) == 1:
                return _query_r[0]["fa_space_id"]
            elif len(_query_r) == 0:
                return None
            else:
                await self._log_cloud(i_stack(),
                                      "warning",
                                      f"search_space(): {len(_query_r)} entries found "
                                      f"with name '{name}' site ID = {siteid} floor ID = {floorid} "
                                      f"please check database for outdated records")
                return _query_r[0]["fa_space_id"]
        except Exception as _exc:
            await self._log_cloud(i_stack(), "error", str(), _exc, format_exc())
            return None

    async def search_user(self, domain: str, name: str) -> Optional[int]:
        """
        Searches user by name, there should be only one, otherwise returns first match and logs warning

        :param domain: FA domain name
        :param name: name to search for
        :return: ID of entry in FA backend
        """
        try:
            async with self._pg.acquire() as _c:
                _query_r = await _c.fetch(self._q["search_user"], domain, name)
            if len(_query_r) == 1:
                return _query_r[0]["fa_id"]
            elif len(_query_r) == 0:
                return None
            else:
                await self._log_cloud(i_stack(),
                                      "warning",
                                      f"search_user(): {len(_query_r)} entries found "
                                      f"with name '{name}' please check database for outdated records")
                return _query_r[0]["fa_id"]
        except Exception as _exc:
            await self._log_cloud(i_stack(), "error", str(), _exc, format_exc())
            return None

    async def _shutdown_clear(self):
        """
        Generic clearing before shutdown overriden
        """
        await self.logout_from_fa()

    @staticmethod
    def _unique_key(domain: str,
                    timetz: datetime,
                    siteid: int,
                    userid: int,
                    floorid: Optional[int] = None,
                    spaceid: Optional[int] = None) -> str:
        """
        Make unique key string for time reg entry
        :param domain: FA domain name
        :param timetz: timezone aware timestamp where time was registered
        :param siteid: FA backed ID of site
        :param userid: FA backed ID of site
        :param floorid: FA backed ID of floor
        :param spaceid: FA backed ID of space
        """
        return hashlib.sha256(dumps({"domain":  domain,
                                     "timetz":  int(timetz.timestamp()),
                                     "siteid":  siteid,
                                     "userid":  userid,
                                     "floorid": floorid,
                                     "spaceid": spaceid
                                     }, ensure_ascii=False).encode("utf-8")).hexdigest()

    async def _app_logins_update(self) -> None:
        """
        Update FA mobile API logins in the PostgreSQL database
        """
        # First get admin records to serve
        await self.get_admin_records()
        # Login to FA mobile API for each admin record and update session_id values into the admin records
        for _admin_record in self._admin_records:
            _fa_app_login_res = await self._fa.faapp_login(_admin_record["fa_domain"],
                                                           _admin_record["fa_usr"],
                                                           _admin_record["fa_pwd"])
            if isinstance(_fa_app_login_res, dict) and "LoggedIn" in _fa_app_login_res.keys():
                if _fa_app_login_res["LoggedIn"] and "session_id" in _fa_app_login_res.keys():
                    await self._log_cloud(i_stack(),
                                          "info",
                                          (f"OK mobile logged in to domain '{_admin_record['fa_domain']}' "
                                           f"user '{_admin_record['fa_usr']}'"))
                    async with self._pg.acquire() as _c:
                        await _c.execute(self._q["upd_admin_session"],
                                         _admin_record["fa_domain"],
                                         _fa_app_login_res["session_id"],
                                         int(time.time()))
                else:
                    await self._log_cloud(i_stack(),
                                          "warning",
                                          (f"refused mobile login to domain '{_admin_record['fa_domain']}' "
                                           f"user '{_admin_record['fa_usr']}': {str(_fa_app_login_res)}"))
            else:
                await self._log_cloud(i_stack(),
                                      "warning",
                                      (f"can not mobile login to domain '{_admin_record['fa_domain']}' "
                                       f"user '{_admin_record['fa_usr']}' "
                                       f"failed API response: {str(_fa_app_login_res)}"))
        # Log success
        await self._log_cloud(i_stack(),
                              "info",
                              f"OK completed app logins update")

    async def _rmq_handle_control(self, msg: aio_pika.abc.AbstractIncomingMessage):
        """
        Consumes messages from the control queue, and performs requested actions

        :param msg: aio-pika incoming message object
        """
        try:
            async with msg.process():
                _command_json = msg.body
            _command = loads(_command_json.decode("utf-8"))
            if isinstance(_command, dict):
                if "cmd" in _command.keys():
                    if _command["cmd"] == "stop":
                        # Stop the cycle
                        self._run_cycle = False
                    else:
                        if _command["cmd"] in self._control_handlers.keys():
                            await self._control_handlers[_command["cmd"]](_command)
                        else:
                            await self._log_cloud(i_stack(),
                                                  "warning",
                                                  (f"unknown command '{_command['cmd']}' received "
                                                   f"from {self._dbname}.{self._servicename}.control queue"))
                else:
                    await self._log_cloud(i_stack(),
                                          "warning",
                                          (f"no 'cmd' key in command object received "
                                           f"from {self._dbname}.{self._servicename}.control queue"))
            else:
                await self._log_cloud(i_stack(),
                                      "warning",
                                      (f"wrong message received "
                                       f"from {self._dbname}.{self._servicename}.control, must be JSON object"))
        except Exception as _exc:
            await self._log_cloud(i_stack(),
                                  "error",
                                  (f"processing command message "
                                   f"from {self._dbname}.{self._servicename}.control queue"),
                                  exc=_exc,
                                  fexc=format_exc())

    async def _log_cloud(self,
                         stack: List[FrameInfo],
                         logl: str,
                         logm: str,
                         exc: Optional[Exception] = None,
                         fexc: Optional[str] = None,
                         logo: Optional[Union[List, Dict]] = None) -> None:
        """
        Logs to stdout / stderr, to RabbitMQ log queue and to logging DB table

        :param stack: result of inspect.stack()
        :param logl: logging level
        :param logm: message to log
        :param exc: optional Exception object
        :param fexc: optional contents format_exc()
        :param logo: optional object to log with dumps, may be list or dict
        """
        _time = time.time()
        # Define log level
        if str(logl).upper() not in ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]:
            _logl = "INFO"
        else:
            _logl = str(logl).upper()
        _loglevel = {
                "CRITICAL": 50,
                "ERROR":    40,
                "WARNING":  30,
                "INFO":     20,
                "DEBUG":    10,
                "NOTSET":   0
                }[_logl]
        # Add the exception name and arguments, if any,
        if isinstance(exc, Exception):
            _logm = f"{str(logm)} {exc.__class__.__name__} {exc.args}"
        else:
            _logm = str(logm)
        # Format object to log
        _logo = logo if isinstance(logo, (list, dict)) else None
        # Need to log if issue with DB
        _need_to_log = False
        # Log to DB
        if self._need_psg:
            try:
                await self._log_db(stack, logl, logm, exc, fexc, logo)
            except Exception as _exc:
                _need_to_log = True
                self._log.error(f"{self.get_stack_str(i_stack())}: {_exc.__class__.__name__} {_exc.args}")
        # Log to stderr / stdout
        if _need_to_log or not self._need_psg:
            _stack_to_log = f"{self.get_stack_str(stack)}: " if isinstance(self.get_stack_str(stack), str) else str()
            if isinstance(fexc, str):
                self._log.log(_loglevel, f"{_stack_to_log}{_logm} {fexc.replace(chr(10), ' ')}")
            else:
                self._log.log(_loglevel, f"{_stack_to_log}{_logm}")

    async def _log_db(self,
                      stack: List[FrameInfo],
                      logl: str,
                      logm: str,
                      exc: Optional[Exception] = None,
                      fexc: Optional[str] = None,
                      logo: Optional[Any] = None) -> bool:
        """
        Logs to stdout / stderr, to RabbitMQ log queue and to logging DB table

        :param stack: result of inspect.stack()
        :param logl: logging level
        :param logm: message to log
        :param exc: optional Exception object
        :param fexc: optional contents format_exc()
        :param logo: optional object to log with dumps, may be list or dict
        :return: False if such same log was already recently added otherwise True
        """
        _perform_db_logging = True
        if self._need_psg:
            # Call stack value to insert
            _stack_insert = self.get_stack_str(stack) if isinstance(self.get_stack_str(stack), str) else None
            # Define log level
            if str(logl).upper() not in ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]:
                _logl = "INFO"
            else:
                _logl = str(logl).upper()
            # Add the exception name and arguments, if any,
            if isinstance(exc, Exception):
                _logm = f"{str(logm)} {exc.__class__.__name__} {exc.args}"
            else:
                _logm = str(logm)
            # Make uniqueness key
            _key = hashlib.sha256(dumps({"level":     _logl.lower(),
                                         "stack":     self.get_stack_str(stack),
                                         "message":   _logm,
                                         "exception": fexc.replace(chr(10), ' ') if isinstance(fexc, str) else None
                                         },
                                        ensure_ascii=False).encode("utf-8")).hexdigest()
            # For warnings, errors and debugs - check if such errors already in the database, not to log duplicates
            if _logl in ["ERROR", "WARNING", "INFO", "DEBUG"]:
                try:
                    async with self._pg.acquire() as _c:
                        _get_logs_by_key_r = await _c.fetch(self._q["get_logs_by_key"],
                                                            _key,
                                                            float(self._config["logdbintvl"]))
                        if len(_get_logs_by_key_r) > 0:
                            _perform_db_logging = False
                except Exception as _exc:
                    self._log.error(f"{self.get_stack_str(i_stack())}: "
                                    f"get_logs_by_key {_exc.__class__.__name__} - {_exc.args}")
            if _perform_db_logging:
                # Format object to log
                _logo = dumps(logo,
                              ensure_ascii=False,
                              default=self._dumps_default) if isinstance(logo, (list, dict)) else None
                # And insert into DB
                try:
                    async with self._pg.acquire() as _c:
                        await _c.execute(self._q["insert_log"],
                                         _logl.lower(),
                                         _logm,
                                         fexc if isinstance(fexc, str) else None,
                                         _logo,
                                         _key,
                                         _stack_insert,
                                         self._servicename)
                except Exception as _exc:
                    self._log.error(f"{self.get_stack_str(i_stack())}: "
                                    f"insert_log {_exc.__class__.__name__} - {_exc.args}")
        return _perform_db_logging



# coding=utf-8
"""
BaseMicroService
"""

import asyncio
import logging
import signal
from contextlib import suppress
from datetime import datetime
from inspect import FrameInfo
from json import dumps, load
from traceback import format_exc
from typing import Any, Dict, List, Optional, Union

import aio_pika
import aiohttp
import asyncpg
import boto3
from aiohttp import web
from botocore.client import BaseClient


# noinspection SqlNoDataSourceInspection
class BaseMicroService:
    """
    Base class to build microservices on top
    """

    def __init__(self,
                 evloop: asyncio.AbstractEventLoop,
                 appname: str,
                 configfile: str,
                 dbname: Optional[str] = None,
                 servicename: Optional[str] = None,
                 tcphost: Optional[str] = None,
                 tcpport: Optional[int] = None,
                 httpmwares: Optional[List] = None,
                 pikattl: int = 60,
                 pikassl: bool = False,
                 customloop: bool = False,
                 psqlport: int = 5432,
                 username: Optional[str] = None,
                 retrymul: float = 1.5,
                 retrymulmaxtime: float = 60):
        """
        Init the base class

        :param evloop: asyncio event loop to run in
        :param appname: name of the whole application
        :param configfile: config file name
        :param dbname: name of database to use
        :param servicename: name of the current service
        :param tcphost: optional TCP host address for web server to listen at
        :param tcpport: optional TCP port number for web server to listen at
        :param httpmwares: list of aiohttp middleware callables as is passed to aiohttp.web.Application
                           middlewares= option
        :param pikattl: RabbitMQ messages ttl
        :param pikassl: use TLS for RabbitMQ connection
        :param customloop: set True if custom main loop is used, in this case event loop must be stopped in the custom
                           loop after asyncio.CancelledError catched
        :param psqlport: optional port number for PostgreSQL server to connect to
        :param username: optional user name for connecting services and to use as linux user name, by default value of
                         dbname is taken as user name
        :param retrymul: optional connection retry time multiplier, each next connection retry interval increases times
                         this
        :param retrymulmaxtime: optional retry interval multiplication max time in seconds, after this time retry
                                interval is not multiplied any more
        """
        # Constants
        self._TTL = pikattl
        self._SSL = pikassl
        # Event loop
        self._evloop = evloop
        self._customloop = customloop
        # Set application name
        self._appname: str = appname
        # Set service name
        if servicename is None:
            self._servicename: str = appname
        else:
            self._servicename: str = servicename
        # Set user name
        if username is None:
            self._username = dbname
        else:
            self._username = username
        # Load config
        self._config: Optional[dict] = None
        with open(configfile, "r") as _config_file_fd:
            self._config = load(_config_file_fd)
        # Load secrets
        with open(f"/home/{self._username}/.secret/.secret", "r") as _secret_file_fd:
            self._secrets = load(_secret_file_fd)
        # Run cycle flag
        self._run_cycle = True
        # Create logging formatter
        self._log_formatter = logging.Formatter(fmt="{asctime}.{msecs:03.0f} {levelname} {process} {name}: {message}",
                                                datefmt="%Y.%m.%d %H:%M:%S",
                                                style="{")
        # Get the logging level from config
        self._base_loglevel = {
            "CRITICAL": 50,
            "ERROR": 40,
            "WARNING": 30,
            "INFO": 20,
            "DEBUG": 10,
            "NOTSET": 0
        }[self._config["loglevel"].upper()]
        # Create service logger
        self._log = logging.getLogger(self._servicename)
        self._log.setLevel(self._base_loglevel)
        self._log_streamhandler = logging.StreamHandler()
        self._log_streamhandler.setFormatter(self._log_formatter)
        self._log.addHandler(self._log_streamhandler)
        self._log.propagate = False
        # Create the generic logger
        # noinspection PyArgumentList
        logging.basicConfig(format="{asctime}.{msecs:03.0f} {levelname} {process} {name}: {message}",
                            datefmt="%Y.%m.%d %H:%M:%S",
                            style="{",
                            level=self._base_loglevel)
        # Connection retry parameters
        self._conn_retry_mul = retrymul
        self._conn_retry_mul_max_time = retrymulmaxtime
        # S3
        self._s3session: Optional[boto3.session.Session] = None
        self._s3client: Optional[BaseClient] = None
        # aiohttp Server
        self._port: Optional[int] = tcpport
        self._host: Optional[str] = tcphost
        if httpmwares is not None:
            self._aiohttp_app: aiohttp.web.Application = web.Application(middlewares=httpmwares)
        else:
            self._aiohttp_app: aiohttp.web.Application = web.Application()
        # aiohttp Runner
        self._aiohttp_app_runner: Optional[aiohttp.web.AppRunner] = web.AppRunner(self._aiohttp_app)
        self._aiohttp_control_site: Optional[aiohttp.web.UnixSite] = None
        self._aiohttp_site: Optional[aiohttp.web.TCPSite] = None
        # aiohttp Routes
        self._routes: List[aiohttp.web.RouteDef] = list()
        self._routes_registered: List[Dict[str, Any]] = list()
        # aiohttp client sessions pool
        self._http_sessions: Dict[str, aiohttp.ClientSession] = dict()
        # RabbitMQ
        self._need_rmq = False
        self._rmq_snd_conn: Optional[aio_pika.connection.Connection] = None
        self._rmq_snd_channel: Optional[aio_pika.channel.Channel] = None
        self._rmq_rcv_conn: Optional[aio_pika.connection.Connection] = None
        self._rmq_rcv_channel: Optional[aio_pika.channel.Channel] = None
        # PostgreSQL
        self._need_psg = False
        self._dbname = dbname
        self._psqlport = psqlport
        # Connection pool
        self._pg: Optional[asyncpg.pool.Pool] = None

    async def _conn_s3(self) -> None:
        """
        Setup the S3 connection to Yandex Object Storage.
        """
        # Create session
        self._s3session = boto3.session.Session(aws_access_key_id=self._config["s3key"],
                                                aws_secret_access_key=self._secrets["s3"])
        # Create client
        self._s3client = self._s3session.client(service_name="s3",
                                                endpoint_url=self._config["s3host"])

    def _init_web_routes(self) -> None:
        """
        Setup the aiohttp web server.
        """
        # Set routes
        _registered = self._aiohttp_app.add_routes(self._routes)
        # Remember routes for logging
        for _route in _registered:
            self._routes_registered.append({"method": _route.method,
                                            "path": _route.resource.canonical
                                            })

    async def _conn_rmq(self) -> None:
        """
        Setup connections and channels to RabbitMQ broker.
        """
        # Default port 5672, if "rmqport" is not provided in config file
        _rmq_port = 5672 if "rmqport" not in self._config.keys() else int(self._config["rmqport"])
        # Connection and channel to send
        self._rmq_snd_conn = await aio_pika.connect_robust(host=self._config["rmqhost"],
                                                           port=_rmq_port,
                                                           client_properties={
                                                               "connection_name": (f"{self._username}"
                                                                                   f".{self._servicename}.snd")
                                                           },
                                                           login=self._username,
                                                           password=self._secrets["rmq"],
                                                           ssl=self._SSL)
        self._rmq_snd_channel: aio_pika.abc.AbstractRobustChannel = await self._rmq_snd_conn.channel(
            publisher_confirms=True)
        # Connection and channel to receive
        self._rmq_rcv_conn = await aio_pika.connect_robust(host=self._config["rmqhost"],
                                                           port=_rmq_port,
                                                           client_properties={
                                                               "connection_name": (f"{self._username}"
                                                                                   f".{self._servicename}.rcv")
                                                           },
                                                           login=self._username,
                                                           password=self._secrets["rmq"],
                                                           ssl=self._SSL)
        self._rmq_rcv_channel: aio_pika.abc.AbstractRobustChannel = await self._rmq_rcv_conn.channel(
            publisher_confirms=True)

    async def _send_rmq(self, qname: str, o: Dict[str, Any], customttl: Optional[int] = None) -> bool:
        """
        Send standard message to RabbitMQ queue

        :param qname: queue name
        :param o: dict object to send
        :param customttl: set custom TTL otherwise service default is used
        """
        try:
            _qmsg = aio_pika.Message(body=dumps(o, ensure_ascii=False, default=self._dumps_time_float).encode("utf-8"),
                                     expiration=(self._TTL if customttl is None else customttl))
            _confirm = await self._rmq_snd_channel.default_exchange.publish(_qmsg,
                                                                            routing_key=qname)
            if not _confirm:
                self._log.warning(f"_send_rmq() not confirmed by broker: {qname}")
                return False
            else:
                return True
        except Exception as _exc:
            self._log.warning(f"_send_rmq() "
                              f"{_exc.__class__.__name__} - {_exc.args}:\n{format_exc().replace(chr(10), ' ')}")
            return False

    @staticmethod
    def get_stack_str(stack: List[FrameInfo]) -> Optional[str]:
        """
        Format call stack into string

        :param stack: result of inspect.stack()
        :return: string representing call stack
        """
        # Compose calling stack string
        if isinstance(stack, list):
            if len(stack) > 0:
                _stack = str()
                for _stack_item in stack:
                    if isinstance(_stack_item, FrameInfo):
                        _stack = f"{_stack_item.function}." + _stack
                return _stack.strip(".")
            else:
                return None
        else:
            return None

    @staticmethod
    def _dumps_default(o: Any) -> Union[str, int, float]:
        """
        Serialize to JSON recognized values
        """
        return str(o)

    @staticmethod
    def _dumps_time_float(o: Any) -> Union[str, int, float]:
        """
        Serialize datetime to JSON as float
        """
        if isinstance(o, datetime):
            return o.timestamp()

    def _log_dump(self, fname: str, o: Any) -> None:
        with open(f"/tmp/{fname}", "w") as _fd:
            _fd.write(dumps(o, ensure_ascii=False, indent=3, default=self._dumps_default))

    def _configure_psg_connections(self, cnames: List[str]) -> None:
        """
        Configure PostgreSQL connections before running setup

        :param cnames: list of connection names to establish to PostgreSQL
        """
        self._pg_conns_names = cnames

    async def _conn_psg(self) -> None:
        """
        Setup connections to PostgreSQL database.
        """
        _break_flag = False
        _retry_interval = 1.0 / self._conn_retry_mul
        while not _break_flag:
            try:
                self._pg = await asyncpg.create_pool(host=self._config["psqlhost"],
                                                     port=self._psqlport,
                                                     user=self._username,
                                                     password=self._secrets["psql"],
                                                     database=self._dbname,
                                                     max_inactive_connection_lifetime=65,
                                                     statement_cache_size=256,
                                                     max_cached_statement_lifetime=0,
                                                     ssl=False,
                                                     min_size=4,
                                                     max_size=64)
                _break_flag = True
            except OSError:
                if _retry_interval * self._conn_retry_mul < self._conn_retry_mul_max_time:
                    _retry_interval *= self._conn_retry_mul
                else:
                    _retry_interval = self._conn_retry_mul_max_time
            if not _break_flag:
                await asyncio.sleep(_retry_interval)
        _break_flag = False
        _retry_interval = 1.0 / self._conn_retry_mul
        while not _break_flag:
            try:
                async with self._pg.acquire() as _c:
                    _pg_version = _c.get_server_version()
                    with suppress(Exception):
                        self._log.info(f"OK PostgreSQL connection pool created ["
                                       f"{not _c.is_closed()}] server version "
                                       f"{_pg_version[0]}.{_pg_version[1]}.{_pg_version[2]} {_pg_version[3]}")
                _break_flag = True
            except OSError:
                if _retry_interval * self._conn_retry_mul < self._conn_retry_mul_max_time:
                    _retry_interval *= self._conn_retry_mul
                else:
                    _retry_interval = self._conn_retry_mul_max_time
            if not _break_flag:
                await asyncio.sleep(_retry_interval)

    async def _shutdown_clear(self) -> None:
        """
        Generic clearing before shutdown virtual - to override in child classes
        """
        raise NotImplementedError("Must override _shutdown_clear()")

    async def _shutdown(self) -> None:
        """
        Safe shutdown
        """
        self._log.info(f"OK received SIGTERM")
        self._evloop.remove_signal_handler(signal.SIGTERM)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        # Stop the cycle if it runs
        self._run_cycle = False
        # Clear custom
        with suppress(Exception):
            await self._shutdown_clear()
        # Shutdown connections
        with suppress(Exception):
            await self._rmq_snd_channel.close()
        with suppress(Exception):
            await self._rmq_rcv_channel.close()
        self._log.info(f"OK closed RabbitMQ channels")
        with suppress(Exception):
            await self._rmq_snd_conn.close()
        with suppress(Exception):
            await self._rmq_rcv_conn.close()
        self._log.info(f"OK closed RabbitMQ connections")
        with suppress(Exception):
            await self._aiohttp_control_site.stop()
        with suppress(Exception):
            await self._aiohttp_site.stop()
        with suppress(Exception):
            await self._aiohttp_app_runner.cleanup()
        with suppress(Exception):
            await self._aiohttp_app.cleanup()
        self._log.info(f"OK closed aiohttp server")
        with suppress(Exception):
            await self._pg.close()
        self._log.info(f"OK closed PostgreSQL database pool and driver")
        # Wait for not finished tasks
        _tasks = [_t for _t in asyncio.all_tasks() if _t is not asyncio.current_task()]
        for _task in _tasks:
            with suppress(asyncio.exceptions.CancelledError):
                _task.cancel()
        self._log.info(f"OK cancelled {len(_tasks)} outstanding tasks")
        if not self._customloop:
            self._evloop.stop()
            self._log.info(f"OK shut down")

    async def setup(self,
                    need_s3: bool = False,
                    need_web: bool = False,
                    need_rmq: bool = False,
                    need_psg: bool = False):
        """
        Setup routine

        :param need_s3: Set true to initiate S3 Storage client
        :param need_web: Set true to initiate web server, routes must already be initiated
        :param need_rmq: Set true to initiate RabbitMQ connections and channels
        :param need_psg: Connect to PostgreSQL database and acquire connection pool
        """
        # Configure web routes
        self._init_web_routes()
        # Log registered routes
        self._log.info(f"OK web server registered routes: {dumps(self._routes_registered, ensure_ascii=False)}")
        # Configure control web server
        await self._aiohttp_app_runner.setup()
        # Connect to the Object Storage
        if need_s3:
            await self._conn_s3()
            self._log.info(f"OK S3 init complete")
        # Initiate web server
        if need_web:
            self._aiohttp_site = web.TCPSite(self._aiohttp_app_runner, host=self._host, port=self._port)
            await self._aiohttp_site.start()
            self._log.info(f"OK TCPSite started")
        # Connect to RabbitMQ
        if need_rmq:
            await self._conn_rmq()
            self._need_rmq = True
            self._log.info(f"OK connected to RabbitMQ")
        # Connect to PostgreSQL database
        if need_psg:
            await self._conn_psg()
            self._need_psg = True
            self._log.info(f"OK connected to PostgreSQL")
        self._log.info(f"OK service {self._servicename} setup complete")
        # Ignore all not needed signals
        for snum in range(15):
            if snum + 1 not in [signal.SIGTERM, signal.SIGKILL]:
                signal.signal(snum + 1, signal.SIG_IGN)
        # And add linux signal handler for SIGTERM
        self._evloop.add_signal_handler(signal.SIGTERM, lambda: asyncio.create_task(self._shutdown()))



# coding=utf-8
"""
samgenericservices
"""

__version__ = "0.9.15"

from typing import Tuple

from .BaseMicroService import (BaseMicroService as BaseMicroService)
from .FAAPIService import (FAAPIService as FAAPIService)
from .IQNFCService import (IQNFCService as IQNFCService)


__all__: Tuple[str, ...] = (
        "BaseMicroService",
        "FAAPIService",
        "IQNFCService",
        )



# coding=utf-8
"""
samgenericservices

py setup.py sdist bdist_wheel
twine upload --config-file E:\twine-config-pypi.pypirc dist/*0.9.15*
"""

from setuptools import setup

setup(name='samgenericservices',
      version='0.9.15',
      packages=['samgenericservices'],
      url='https://t.me/samthesuperhero',
      license='Creative Commons Attribution 4.0 International',
      author='Pavel A. Fomin',
      author_email='fomin-p@yandex.ru',
      description='Classes and functions to use for development of intregration software',
      install_requires=['aiohttp>=3.8.4',
                        'aio_pika>=9.0.5',
                        'aiormq>=6.7.4',
                        'asyncpg>=0.27.0',
                        'boto3>=1.22.8',
                        'botocore>=1.25.8',
                        'bs4>=0.0.1',
                        'multidict>=6.0.4',
                        'viledatools>=1.0.34'],
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Other Environment',
                   'License :: Other/Proprietary License',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.8',
                   'Topic :: Utilities'],
      zip_safe=False,
      python_requires='>=3.8'
      )
